@@ -16,14 +16,14 @@ import tempfile
 import time
 import traceback
 from contextlib import contextmanager
-from typing import Dict, List, Optional, Set, Tuple
+from typing import Callable, Dict, List, Optional, Set, Tuple
 from urllib.parse import urlsplit
 
 import termcolor
 import yaml
 from ansi2html import Ansi2HTMLConverter
 from dateutil import parser
-from docker.models.resource import Model
+from docker.models.containers import Container
 from six import BytesIO
 from zuper_commons.text import box
 from zuper_commons.types import ZException
@@ -48,12 +48,14 @@ from duckietown_challenges.constants import (
     ENV_CHALLENGE_STEP_NAME,
 )
 from duckietown_challenges.rest_methods import (
+    dtserver_job_heartbeat,
     dtserver_report_job,
     dtserver_work_submission,
     get_challenge_description,
     SubmissionID,
 )
 from duckietown_challenges.utils import friendly_size, indent
+from duckietown_docker_utils.docker_run import get_developer_volumes
 from . import __version__, __version__ as version_dcr
 from .env_checks import (
     check_docker_environment,
@@ -101,7 +103,7 @@ def dt_challenges_evaluator():
 
 
 # noinspection PyBroadException
-def dt_challenges_evaluator_():
+def dt_challenges_evaluator_(args: List[str] = None, token: str = None):
     from duckietown_challenges.col_logging import setup_logging
 
     setup_logging()
@@ -113,10 +115,10 @@ def dt_challenges_evaluator_():
 Usage:
 
 """
-    args = sys.argv[1:]
-    if args and args[0].startswith("dt-"):
-        elogger.debug("removing first arg from %s" % args)
-        args = args[1:]
+    # args = sys.argv[1:]
+    # if args and args[0].startswith("dt-"):
+    #     elogger.debug("removing first arg from %s" % args)
+    #     args = args[1:]
     prog = "dt-challenges-evaluator"
     parser = argparse.ArgumentParser(prog=prog, usage=usage)
     parser.add_argument("--version", action="store_true", default=False)
@@ -177,11 +179,12 @@ Usage:
     do_upload = not parsed.no_upload
     delete = not parsed.no_delete
     reset = parsed.reset
-    evaluator_name = parsed.name or "p-%s" % os.getpid()
+    evaluator_name = (parsed.name or "noname") + "-" + "%s" % socket.gethostname() + "-" + "%s" % os.getpid()
     machine_id = parsed.machine_id or socket.gethostname()
     allow_host_network = False
 
-    token = get_token_from_shell_config()
+    if token is None:
+        token = get_token_from_shell_config()
 
     args = dict(
         do_upload=do_upload,
@@ -480,6 +483,27 @@ def go_(
     else:
         cd = None
 
+    def heartbeat():
+        res = None
+        try:
+            res = dtserver_job_heartbeat(
+                token,
+                job_id=job_id,
+                machine_id=machine_id,
+                process_id=process_id,
+                evaluator_version=evaluator_version,
+                impersonate=impersonate,
+            )
+
+            abort = res["abort"]
+            why = res["why"]
+        except:
+            elogger.warning(traceback.format_exc(), res=res)
+        else:
+            if abort:
+                msg = f"The server told us to abort the job because: {why}"
+                raise KeyboardInterrupt(msg)
+
     cr, uploaded = get_cr(
         res,
         tmpdir,
@@ -494,6 +518,7 @@ def go_(
         use_ipfs=use_ipfs,
         server_host=server_host,
         cd=cd,
+        heartbeat=heartbeat,
     )
 
     msg = "Reporting status %s for job %s submission %s.\n\n%s\n\n%s\n\n%s" % (
@@ -551,6 +576,7 @@ def get_cr(
     use_ipfs: bool,
     server_host: str,
     cd: Optional[ChallengeDescription],
+    heartbeat: Optional[Callable] = None,
 ) -> Tuple[ChallengeResults, List]:
     # noinspection PyBroadException
     try:
@@ -618,7 +644,9 @@ def get_cr(
             with setup_logging(own_logs_dir):
                 challenge_parameters_ = EvaluationParameters.from_yaml(res["challenge_parameters"])
 
-                project = "job%s-%s" % (job_id, random.randint(1, 1000000))
+                rnd = random.randint(1, 1000000)
+
+                project = f"{evaluator_name}-job{job_id}-{rnd}"
                 if not do_upload:
                     aws_config = None
 
@@ -638,6 +666,7 @@ def get_cr(
                     use_ipfs=use_ipfs,
                     server_host=server_host,
                     require_scores=require_scores,
+                    heartbeat=heartbeat,
                 )
         finally:
             if do_upload:
@@ -707,6 +736,7 @@ def run_single(
     require_scores: Set[str],
     allow_host_network: bool = False,
     use_ipfs: bool = False,
+    heartbeat: Optional[Callable] = None,
 ):
     """
 
@@ -725,6 +755,8 @@ def run_single(
     :param require_scores: Set of scores required for this step to be valid
     :param allow_host_network:
     :param use_ipfs:
+    :param heartbeat:
+
     :return:
     """
     prepare_dir(wd, aws_config, steps2artefacts, use_ipfs=use_ipfs)
@@ -792,7 +824,13 @@ def run_single(
         services_names = list(challenge_parameters.services)
         # print('services: %s' % services_names)
         cr = run_one(
-            wd, project, do_pull, services=services_names, monitor=tomonitor[0], timeout_sec=timeout_sec,
+            wd,
+            project,
+            do_pull,
+            services=services_names,
+            monitor=tomonitor[0],
+            timeout_sec=timeout_sec,
+            heartbeat=heartbeat,
         )
 
         write_logs(wd, project, services=config["services"])
@@ -878,13 +916,21 @@ def teminate_workers():
 
 
 # noinspection PyBroadException
-def run_one(wd: str, project: str, do_pull: bool, monitor: ContainerName, services, timeout_sec: float):
+def run_one(
+    wd: str,
+    project: str,
+    do_pull: bool,
+    monitor: ContainerName,
+    services,
+    timeout_sec: float,
+    heartbeat: Optional[Callable] = None,
+):
     client = check_docker_environment()
 
     # noinspection PyBroadException
     try:
         if do_pull:
-            elogger.info("pulling containers")
+            elogger.info("pulling containers...")
             cmd = ["pull"]  # '--ignore-pull-failures'
             run_docker(wd, project, cmd)
 
@@ -923,8 +969,27 @@ def run_one(wd: str, project: str, do_pull: bool, monitor: ContainerName, servic
         client = check_docker_environment()
 
         for service_name, container_id in service2id.items():
-            container: Model = client.containers.get(container_id)
+            container: Container = client.containers.get(container_id)
             container.reload()
+            #
+            # cpu.cfs_quota_us: the total available run-time within a period (in microseconds)
+            # cpu.cfs_period_us: the length of a period (in microseconds)
+            # cpu.stat: exports throttling statistics [explained further below]
+
+            if "solution" in service_name:
+                if False:
+                    # every second
+                    cpu_period = 1000 * 1000
+                    # get 100% cpu
+                    cpu_quota = 1000 * 1000
+
+                    container.update(cpu_period=cpu_period, cpu_quota=cpu_quota)
+                    elogger.info(
+                        "Applying CPU quota",
+                        service_name=service_name,
+                        cpu_period=cpu_period,
+                        cpu_quota=cpu_quota,
+                    )
 
             # noinspection PyUnresolvedReferences
             ports = container.ports
@@ -942,6 +1007,8 @@ def run_one(wd: str, project: str, do_pull: bool, monitor: ContainerName, servic
 
         exited = []
         while True:
+            if heartbeat:
+                heartbeat()
             running = []
             for service_name, container_id in service2id.items():
                 if service_name in exited:
@@ -1057,7 +1124,7 @@ def prepare_dir(wd, aws_config, steps2artefacts, use_ipfs: bool):
 
 
 def get_config(
-    wd,
+    wd: str,
     fd,
     challenge_parameters_,
     solution_container: str,
@@ -1156,24 +1223,30 @@ def get_config(
     # add volumes
     # wd = os.path.join(os.getcwd(), wd)
     # fd = os.path.join(os.getcwd(), fd)
-    volumes = [
-        f"{wd}:/challenges",
-        f"{fd}:/fifos",
-        f"{dir_fake_home_host}:{dir_home_guest}",
-    ]
+    # "/Volumes/work/DT/dt-env/dt-env-developer/src/aido-analyze/src/aido_analyze": {
+    #             "bind": "/usr/local/lib/python3.8/dist-packages/aido_analyze",
+    #             "mode": "ro"
+    #         }
+    volumes0 = {}
+    volumes0[wd] = {"bind": "/challenges", "mode": "rw"}
+    volumes0[fd] = {"bind": "/fifos", "mode": "rw"}
+    volumes0[dir_fake_home_host] = {"bind": dir_home_guest, "mode": "rw"}
+
     if use_ipfs:
         if os.path.exists("/ipfs"):
-            volumes.append("/ipfs:/ipfs:ro")
+            volumes0["ipfs"] = {"bind": "/ipfs", "mode": "ro"}
         else:
             msg = "/ipfs mount point not found"
             raise IPFSException(msg)
 
     if debug_volumes:
-        data = yaml.load(open(debug_volumes).read(), Loader=yaml.SafeLoader)
-        volumes.extend(data["volumes"])
+        volumes0.update(get_developer_volumes())
+        # data = yaml.load(open(debug_volumes).read(), Loader=yaml.SafeLoader)
+        # volumes.extend(data["volumes"])
 
     for service in config["services"].values():
         assert "volumes" not in service
+        volumes = [f'{k}:{v["bind"]}:{v["mode"]}' for k, v in volumes0.items()]
         service["volumes"] = copy.deepcopy(volumes)
 
     # elogger.info('Now:\n%s' % safe_yaml_dump(config))
@@ -1278,10 +1351,10 @@ class Tee:
             return f.read()
 
 
-def run_docker(cwd: str, project: str, cmd0: List[str], get_output=False) -> bytes:
+def run_docker(cwd: str, project: str, cmd0: List[str], get_output: bool = False) -> bytes:
     cmd0 = ["docker-compose", "-p", project] + cmd0
     # elogger.info('Running:\n\t%s' % " ".join(cmd0) + '\n\n in %s' % cwd)
-    # elogger.debug('Running: %s' % " ".join(cmd0))
+    elogger.debug("Running: %s" % " ".join(cmd0))
 
     d = tempfile.mkdtemp()
 
@@ -1297,7 +1370,8 @@ def run_docker(cwd: str, project: str, cmd0: List[str], get_output=False) -> byt
             return subprocess.check_output(cmd0, cwd=cwd, stderr=sys.stderr)
         else:
             # noinspection PyTypeChecker
-            subprocess.check_call(cmd0, cwd=cwd, stdout=tee_stdout, stderr=tee_stderr)
+            # subprocess.check_call(cmd0, cwd=cwd, stdout=tee_stdout, stderr=tee_stderr)
+            subprocess.check_call(cmd0, cwd=cwd, stdout=sys.stdout, stderr=sys.stderr)
     except subprocess.CalledProcessError as e:
         msg = "Could not run %s:\n\n %s" % (cmd0, indent(e, "  >  "))
         msg += "\n\n%s" % indent(tee_stdout.getvalue(), "stdout | ")
